# Copyright (C) 2023 Callum Dickinson
#
# Buildarr is free software: you can redistribute it and/or modify it under the terms of the
# GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
#
# Buildarr is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Buildarr.
# If not, see <https://www.gnu.org/licenses/>.


"""
Sonarr plugin configuration utility classes and functions.
"""


from __future__ import annotations

from typing import TYPE_CHECKING

from buildarr.config import ConfigBase
from pydantic import ConstrainedStr

if TYPE_CHECKING:
    from ..secrets import SonarrSecrets

    class SonarrConfigBase(ConfigBase[SonarrSecrets]):
        ...

else:

    class SonarrConfigBase(ConfigBase):
        ...


class TraktAuthUser(ConstrainedStr):
    """
    Constrained string type to make the Trakt auth user case-insensitive.
    """

    min_length = 1
    to_lower = True
